# This is 'NakaMoo', a converter of courseNavi csv to Moodle xml.
# (Copyright) 2020 Osamu NAKAMURA
#
# Usage:
# - Make sure you have the following three files.
#   These must be in the same folder.
#     - This python program (.py).
#     - Windows batch file (.bat) (same name as the .py program).
#     - CourseNavi csv file.
# - Drag and drop the courseNavi csv file to the .bat file.
#   .xml file will then be generated.
# - If you face an error, please check _log.txt.
#
# Notes:
# - Developped with python 3.8.
# - Based on Moodle's xml format and tags at the moment of 2020-Apr.
# - This program can handle the followings:
#     自由記述／単一選択／複数選択/穴埋め（文字列型）.
#   The other form of questions, if there are any, are ignored.
# - 1st argument (argv[1]): input file.
# - Put all the three files in the same folder.
#   The batch file may not recognize input files in other folders.
# - The batch file is only needed if yo use drag-and-drop.
# - "timeout=xxx"(sec) may be set in the batch file for debugging.
#   It may be removed for general use.
#
# ver.002     2020-04-12
#   - First release to the public.
# ver.003     2020-04-18
#   - Minor revision.
#   - <responsefieldlines>15 -> 7 in 自由記述
# ver.004-005 2020-04-19
#   The followings enabled:
#   - Transferring cNavi's grouping of sections and random questions
#     to Moodle.
#   - Changing category name in importing ('lab_cat' variable)
#   - Option to Add the question title used in cNavi to the beginning
#     of the question text on Moodle (flag_qttl)
#   - Cloze questions (穴埋め問題)
# ver.006     2020-04-20
#   - 'description'（単なる説明で解答不要のもの） that are put at
#     the beginning of new sections included.
# ver.007     2020-04-25
#   - Minor revision on the treatment of spaces in csv
#     (split() -> split('\n')).
#   - Size of cloze question blakns made adjustable.
# ver.008     2020-04-26
#   - Enabled Capital/small letter control in cloze questions according
#     to courseNavi setting.
#   - Sub-category policy revised (flag_sbct=2 added).
# ver.009     2020-05-02
#   - Zenkaku/Hankaku control partially enabled.
#   - Enabled flag_mgrd option on the treatment of partial grades
#     in multiple-choice questions.
# ver.010     2020-05-03
#   - 選択形式問題 (a set of questions that examinees can choose which to
#     answer to) covered.
# ver.012     2020-05-10
#   - flag_qttl=2 added.
# ver.013     2020-05-16
#   - Switching to flag_qttl=1 enabled if "設問タイトル" of the csv header
#     is replaced by "all_設問タイトル". It is more convenient to specify
#     the flag in the input csv file rather than in this program.
# ver.014     2020-10-05
#   - NUMERICAL in cloze enabled.

import os
import sys
import csv

import random, string
    # These two are used by chr_rnd().

# ==============================================================================
# Function to create a random string that has a length of given n.
# Source: @Scstechr on Qiita
# https://qiita.com/Scstechr/items/c3b2eb291f7c5b81902a

def chr_rnd(n):
    c_rnd = [random.choice(string.ascii_letters + string.digits) for i in range(n)]
    return ''.join(c_rnd)
# ==============================================================================
# Function to transform zenkaku <--> hankaku (Only for 数字・記号・アルファベット)
# Source: @YuukiMiyoshi on Qiita
# https://qiita.com/YuukiMiyoshi/items/6ce77bf402a29a99f1bf

def zen2han(z_chr):
    h_chr = z_chr.translate(str.maketrans({chr(0xFF01 + i): chr(0x21   + i) for i in range(94)}))
    return  h_chr

def han2zen(h_chr):
    z_chr = h_chr.translate(str.maketrans({chr(0x0021 + i): chr(0xFF01 + i) for i in range(94)}))
    return  z_chr
# ==============================================================================


##########
########## settings/initial process
##########

flag_nans =  0
    # Hide or show the number of correct answers in multiple-choice questions after answering.
    # Default=0 [ 0: hide / 1: show ]
flag_sbct =  2
    # Use sub-categories to put the questions into when importing to Moodle?
    # Default=2 [ 0: No   / 1: Full-use  / 2: Divide only random-form questions ]
flag_qttl =  2
    # Embed the title of the question used in cNavi in Moodle?
    # Default=2 [ 0: No   / 1: Yes  / 2: Yes only for section/random-Q breaks ]
    # The setting will be changed to flag_qttl=1 if "設問タイトル" of the csv header
    # is changed to "all_設問タイトル".
flag_wdth =  1
    # Transfer cNavi's setting for the width of cloze question blanks to Moodle?
    # Default=1 [ 0: No   / 1: Yes  ]
fwdth     =  1.0
    # Factor to adjust the width of the blanks in cloze questions on Moodle when flag_wdth==1.
    # If the width of the blanks are too narrow, try 1.5 or larger values.
    # If it is too wide, try 0.6 or smaller values.
    # Default=1.0
flag_znhn =  0
    # Transfer cNavi's setting for Zenkaku-Hankaku treatment of cloze questions?
    # Default=0 [ 0: No   / 1: Yes  ]
flag_mgrd =  0
    # On the grading of a false answer choice in multiple-choice questions.
    # Default=0 [ 0: Give minus point that is equal to those for correct answer
    #             choices but for the sign / Not 0: Give minus -100%,
    #             i.e., choosing a wrong answer results in 0 ]
lab_cat   = 'cNavi'
    # Category name where questions are moved to in Moodle.
    # Sub-categories are to be created under this category.
    # Default='cNavi'

f_in_name = sys.argv[1]
f_ot_name = os.path.splitext(f_in_name)[0] + '_xml.xml'
f_lg_name = os.path.splitext(f_in_name)[0] + '_log.txt'
    # input/output file names.
    # argv[1] is the input file.

f_in = open( f_in_name,  'r' )
f_ot = open( f_ot_name,  'w', encoding = 'utf_8' )
f_lg = open( f_lg_name,  'w', encoding = 'utf_8' )
    # open the files

if os.path.splitext(f_in_name)[1] != '.csv':
    f_lg.write('\n')
    f_lg.write(' Error: Your input file is not csv.'             +'\n')
    f_lg.write('        Your input file = ' + f_in_name          +'\n')
    sys.exit()
    # check the file extension


##### header part of the output

f_ot.write('<?xml version=\"1.0\" encoding=\"UTF-8\"?>'      +'\n')
f_ot.write('<quiz>'                                          +'\n')
f_ot.write('\n')

if flag_sbct == 0:
    f_ot.write('  <question type=\"category\">'                  +'\n')
    f_ot.write('    <category>'                                  +'\n')
    f_ot.write('      <text>$course$/top/' + lab_cat + '</text>' +'\n')
    f_ot.write('    </category>'                                 +'\n')
    f_ot.write('    <info format=\"html\">'                      +'\n')
    f_ot.write('      <text></text>'                             +'\n')
    f_ot.write('    </info>'                                     +'\n')
    f_ot.write('    <idnumber></idnumber>'                       +'\n')
    f_ot.write('  </question>'                                   +'\n')
    f_ot.write('\n')
    # If sub-categories are not used, all the questions are put into 'lab_cat' category.


##########
########## main process
##########

reader = csv.reader(f_in)

i_sc   =  0
n_qn   =  0
i_qn   =  0
n_rn   =  0
i_rw   =  0
id_fst = 'not_yet'
        # i_sc: current section  number (section  ID)
        # n_qn: total number of questions in the current section
        # i_qn: current question number (question ID) in the current section
        # n_rn: total number of questions to answer if they are given randomly
        # i_rw: current row number of the input csv file in processing
        # id_fst: A category has been printed to the output file?
        #       [ not_yet / processed ]

for row in reader:
        # Main loop for rows in the input csv file.

    i_rw += 1

    if len(row) != 25:
        f_lg.write('\n')
        f_lg.write(' Error: Line number of the input csv = '            + str(i_rw)               + '\n' )
        f_lg.write('        The number of columns is inconsistent with the correct cNavi format.' + '\n' )
        f_lg.write('        The number of columns in the input file = ' + str(len(row))           + '\n' )
        sys.exit()
        # check the number of columns in the input cNavi csv

    id_cmp = 'no'
        # initialize id_cmp:
        # The process of the current row is [ yes: completed / no: not completed ].
    id_nws = 'no'
        # initialize id_nws:
        # The current row is [ yes: the beginning of a new section / no ].
    if row[2] == 'all_設問タイトル':
        flag_qttl = 1
        # Override flag_qttl setting if "設問タイトル" of the csv header is
        # replaced by "all_設問タイトル".
        f_lg.write('\n')
        f_lg.write(' flag_qttl is set to 1 following the csv header setting.' + '\n' )


    ##########
    ########## category and label of the question
    ##########

    if  n_qn  > 0:
    # Case the current row is under a clearly-defined section.
        i_qn += 1
        if  i_qn > n_qn:
            # Case the current section has ended in the previous row.
            # The variables are then reset to proceed to a new section.
            n_qn    = 0
            i_qn    = 0
            n_rn    = 0
            i_sc   += 1
            id_nws  = 'yes'
        elif row[0] == '11' or row[0] == '12' or row[0] == '13':
            # Case a new section comes without finishing the previous one.
            f_lg.write('\n')
            f_lg.write(' Error: Number of questions in the current section inonsistent.' + '\n')
            sys.exit()

    if  row[0] == '11' or row[0] == '12' or row[0] == '13':
    # Case a new section begins.
    # This check should not be `elif` of n_qn>0 but a new 'if'.
        i_sc        = int(row[1])
        n_qn        = int(row[9])
        i_qn        = 0
        id_nws      = 'yes'

        lab_q       = '{:02d}-{:0>2s}-newsec'.format( i_sc, '0' )
            # Label for the beginning of new sections that is clearly defined in the
            # input csv file. Refer to another part for the detail of the format.

        if   row[0] == '12':
            n_rn    = int(row[10])
            id_rn   = 'rnd'
            # Case the section consists of random-form questions.
        elif row[0] == '13':
            n_rn    = int(row[10])
            id_rn   = 'slc'
            # Case the section consists of choice-form questions.
        else:
            n_rn    = 0
            # Case the section consists of normal      questions.

    else:
    # Case the current row gives a question.
    # A label (lab_q) is given for the question here.

        if  n_rn    > 0:
            lab_q   = '{:02d}-{:0>2s}'.format( i_sc, row[1] ) + '-' + id_rn + '{:02d}'.format( n_rn )
            # Label for random-/choice-form questions
        else:
            lab_q   = '{:02d}-{:0>2s}'.format( i_sc, row[1] )
            # Label for normal      questions
            # Note i_sc/n_rn are integers while row[1] is a string.
            # :02d  -> 2 letters for integer(d), varied with '0' for blanks.
            # :0>2s -> 2 letters for string (s), varied with '0' for blanks.


    ##########
    ########## leave sub-category label in the output
    ##########

        #####
        ##### print sub-category name
        #####

    if flag_sbct == 1 or flag_sbct == 2:
        if ( id_nws == 'yes' or
            ( id_fst == 'not_yet' and row[0] != '設問スタイル' and row[0] != '' ) ):
            # flag_sbct==1 --> Use sub-categories.
            # flag_sbct==2 --> Use sub-categories only for random-form questions.
            # The sub-category label is put in the output file when a new section begins
            # (i.e., id_nws==1). Another case a sub-category label is needed is the start
            # point of the csv file when the first question comes (id_fst==not_yet and
            # the current row has a question).

            f_ot.write('  <question type=\"category\">'                  +'\n')
            f_ot.write('    <category>'                                  +'\n')

            if   flag_sbct == 1:
                f_ot.write('      <text>$course$/top/' + lab_cat + '/' + 'sec{:02d}'.format(i_sc) + '</text>' +'\n')
            elif flag_sbct == 2:
                f_ot.write('      <text>$course$/top/' + lab_cat                                  + '</text>' +'\n')
            else:
                f_lg.write(' Error 256\n')
                sys.exit()
            # If flag_sbct==1, a new sub-category label is given for every start of new section.
            # If flag_sbct==2, the source category name is given for the same case, except for
            # random-form questions.

            f_ot.write('    </category>'                                 +'\n')
            f_ot.write('    <info format=\"html\">'                      +'\n')
            f_ot.write('      <text></text>'                             +'\n')
            f_ot.write('    </info>'                                     +'\n')
            f_ot.write('    <idnumber></idnumber>'                       +'\n')
            f_ot.write('  </question>'                                   +'\n')
            f_ot.write('\n')

            id_fst = 'processed'
            # Once a category is printed, id_fst is toggled to 'processed'.


        #####
        ##### print the question sentence
        #####

        # If row[0]==11 or 12 or 13, the row contains question sentences.
        # They are put to the output here.

    if row[0] == '11' or row[0] == '12' or row[0] == '13':

        if flag_qttl == 1 or flag_qttl == 2: qttl = '<p><b>' + row[2] + '</b><br></p>'
        else:                                qttl = ''
            # if flag_qttl is set, the title of the quetsion that was used
            # in cNavi is embedded at the beginning of the question text.

        f_ot.write('<!-- question: row_' + str(i_rw) + '  -->'        +'\n')
        f_ot.write('  <question type=\"description\">'                +'\n')
        f_ot.write('    <name>'                                       +'\n')
        f_ot.write('      <text>' + lab_q + '-' + row[2] + '</text>'  +'\n')
        f_ot.write('    </name>'                                      +'\n')
        f_ot.write('    <questiontext format=\"html\">'               +'\n')
        f_ot.write('      <text><![CDATA[' + qttl + '<p>' + row[3] + '<br></p>]]></text>' +'\n')
        f_ot.write('    </questiontext>'                              +'\n')
        f_ot.write('    <generalfeedback format=\"html\">'            +'\n')
        f_ot.write('      <text></text>'                              +'\n')
        f_ot.write('    </generalfeedback>'                           +'\n')
        f_ot.write('    <defaultgrade>0.0000000</defaultgrade>'       +'\n')
        f_ot.write('    <penalty>0.0000000</penalty>'                 +'\n')
        f_ot.write('    <hidden>0</hidden>'                           +'\n')
        f_ot.write('    <idnumber></idnumber>'                        +'\n')
        f_ot.write('  </question>'                                    +'\n')
        f_ot.write('\n')

        id_cmp     = 'yes'
        # id_cmp is changed to 'yes' (i.e., the row has been processed)


        #####
        ##### print one more sub-category name
        ##### if the questions in the following rows are in 'random'-/'choice'-form
        #####

        # To make the random/choice-form questions grouped in a unique sub-category on Moodle,
        # one more sub-category for them (*_rnd/*_slc) is created here.

    if row[0] == '12' or row[0] == '13':
        # Case the following rows of csv are random-form questions

        f_ot.write('  <question type=\"category\">'                  +'\n')
        f_ot.write('    <category>'                                  +'\n')
        f_ot.write('      <text>$course$/top/' + lab_cat + '/' + 'sec{:02d}'.format(i_sc) + '_' + id_rn + '</text>' +'\n')
        f_ot.write('    </category>'                                 +'\n')
        f_ot.write('    <info format=\"html\">'                      +'\n')
        f_ot.write('      <text></text>'                             +'\n')
        f_ot.write('    </info>'                                     +'\n')
        f_ot.write('    <idnumber></idnumber>'                       +'\n')
        f_ot.write('  </question>'                                   +'\n')
        f_ot.write('\n')


    ##########
    ########## 自由記述問題
    ##########

    if row[0] == '4':

        if flag_qttl == 1:   qttl = '<p><b>' + row[2] + '</b><br></p>'
        else:                qttl = ''
            # if flag_qttl is set, the title of the quetsion that was used
            # in cNavi is embedded onto the beginning of the question text.

        f_ot.write('<!-- question: row_' + str(i_rw) + '  -->'        +'\n')
        f_ot.write('  <question type=\"essay\">'                      +'\n')
        f_ot.write('    <name>'                                       +'\n')
        f_ot.write('      <text>' + lab_q + '-' + row[2] + '</text>'  +'\n')
        f_ot.write('    </name>'                                      +'\n')
        f_ot.write('    <questiontext format=\"html\">'               +'\n')
        f_ot.write('      <text><![CDATA[' + qttl + '<p>' + row[3] + '<br></p>]]></text>' +'\n')
        f_ot.write('    </questiontext>'                              +'\n')
        f_ot.write('    <generalfeedback format=\"html\">'            +'\n')
        f_ot.write('      <text>' + row[13] + '</text>'               +'\n')
        f_ot.write('    </generalfeedback>'                           +'\n')
        f_ot.write('    <defaultgrade>' + row[4] + '</defaultgrade>'  +'\n')
        f_ot.write('    <penalty>0.0000000</penalty>'                 +'\n')
        f_ot.write('    <hidden>0</hidden>'                           +'\n')
        f_ot.write('    <idnumber></idnumber>'                        +'\n')
        f_ot.write('    <responseformat>editor</responseformat>'      +'\n')
        f_ot.write('    <responserequired>1</responserequired>'       +'\n')
#       f_ot.write('    <responsefieldlines>15</responsefieldlines>'  +'\n')
        f_ot.write('    <responsefieldlines>7</responsefieldlines>'   +'\n')
        # the number of lines in the box can be edited here.
        f_ot.write('    <attachments>0</attachments>'                 +'\n')
        f_ot.write('    <attachmentsrequired>0</attachmentsrequired>' +'\n')
        f_ot.write('    <graderinfo format=\"html\">'                 +'\n')
        f_ot.write('      <text></text>'                              +'\n')
        f_ot.write('    </graderinfo>'                                +'\n')
        f_ot.write('    <responsetemplate format=\"html\">'           +'\n')
        f_ot.write('      <text></text>'                              +'\n')
        f_ot.write('    </responsetemplate>'                          +'\n')
        f_ot.write('  </question>'                                    +'\n')
        f_ot.write('\n')

        id_cmp = 'yes'


    ########## 単一選択問題　または　
    ########## 複数選択問題で部分点あり　または
    ########## 複数選択問題で部分点なし　の場合

    if row[0] == '1' or row[0] == '2':

        if   row[0] == '1':
            id_proc  =    'singl'
        elif row[0] == '2' and row[5] == '1':
            id_proc  =    'm_prt'
        elif row[0] == '2' and row[5] != '1':
            id_proc  =    'm_a_n'
        else:
            f_lg.write('\n')
            f_lg.write(' Error: Unexpected case occurred.' + '\n')
            sys.exit()
        # Identify which case the question is:
        # id_proc = singl: 単一選択問題
        #           m_prt: 複数選択で部分点あり
        #           m_a_n: 複数選択で部分点なし (all-or-nothing)

        choice = row[7].split('\n')
        ans    = row[8].split('\n')
        # de-compose the answer choices and correct answers. This works!

        c_cor  = [0] * len(choice)
        # initialize c_cor as 0.
        # c_cor[]: The answer choice is [ 0: not correct / 1: correct ].
        i_c    =  -1
        for c_now in choice:
            i_c += 1
            for a_now in ans:
                if  c_now == a_now:
                    c_cor[i_c]  = 1
        # find which choice is the correct answer and record it in c_cor.

        n_cans = sum( x==1 for x in c_cor )
        n_fans = sum( x==0 for x in c_cor )
        # These count the num. of c_cor==1/0 each.
        # n_cans: number of correct answers in the answer choices.
        # n_fans: number of false   answers in the answer choices.

        if n_cans == 0:
            f_lg.write('\n')
            f_lg.write(' Error: Line number of the input csv = ' + str(i_rw)  + '\n')
            f_lg.write('        Correct answer in multiple-choice not found.' + '\n')
            sys.exit()
        # check if the correct answer is found in the choices.

        if   id_proc == 'singl' or id_proc == 'm_a_n':
            tot_grd   = float(row[4])
        elif id_proc == 'm_prt':
            tot_grd   = float(row[4]) * n_cans
            # cNavi says 複数選択式で部分配点のとき、その設問の配点は「配点×正解の選択肢数」
        else:
            f_lg.write(' Error 433\n')
            sys.exit()
        # determine tot_grd, the total grade (point) of this question.

        #####
        ##### Determine grade (point) of each answer choice.
        #####

        c_grd = [0] * len(choice)
        # initialize c_grd[] as 0.
        # c_grd[]: grade (point) of each answer choice.
        i_c     = -1
        for c_now in choice:
            i_c += 1

            if   c_cor[i_c] == 1:
            # case the ansewr choice is correct

                if    id_proc == 'singl' or id_proc == 'm_a_n':
                    c_grd[i_c] =  100.0
                    # all-or-nothing mode uses '100' even for multiple correct answers.
                elif  id_proc == 'm_prt':
                    c_grd[i_c] =  100.0 / n_cans
                    # The total of correct answers have to be 100% when partial point is allowed.
                else:
                    f_lg.write(' Error 357\n')
                    sys.exit()

            elif c_cor[i_c] == 0:
            # case the answer choice is false

                if    id_proc == 'singl' or id_proc == 'm_a_n':
                    c_grd[i_c] =    0.0
                elif  id_proc == 'm_prt':
                    if    flag_mgrd == 0:   c_grd[i_c] = -100.0 / n_cans
                    else:                   c_grd[i_c] = -100.0
                    # A minus point have to be set for this case.
                    # If this is set to 0.0, choosing all the choices would give the full point.
                else:
                    f_lg.write(' Error 471\n')
                    sys.exit()

            else:
                    f_lg.write(' Error 475\n')
                    sys.exit()


        #####
        ##### output xml (the first section)
        #####

        f_ot.write('<!-- question: row_' + str(i_rw) + '  -->'       + '\n')

        if id_proc == 'm_a_n':
            f_ot.write('  <question type=\"multichoiceset\">'            + '\n')
            # case all-or-nothing multiple choice
        else:
            f_ot.write('  <question type=\"multichoice\">'               + '\n')
            # case normal multiple choice

        f_ot.write('    <name>'                                      + '\n')
        f_ot.write('      <text>' + lab_q + '-' + row[2] + '</text>' + '\n')
        f_ot.write('    </name>'                                     + '\n')
        f_ot.write('    <questiontext format=\"html\">'              + '\n')


        if flag_qttl == 1: qttl = '<p><b>' + row[2] + '</b><br></p>'
        else:              qttl = ''
            # if flag_qttl is set, the title of the quetsion that was used
            # in cNavi is embedded at the beginning of the question text.

        f_ot.write('      <text><![CDATA[' + qttl + '<p>' + row[3] + '<br></p>]]></text>' + '\n')
        f_ot.write('    </questiontext>'                             + '\n')
        f_ot.write('    <generalfeedback format=\"html\">'           + '\n')
        f_ot.write('      <text>' + row[13] + '</text>'              + '\n')
        f_ot.write('    </generalfeedback>'                          + '\n')
        f_ot.write('    <defaultgrade>' + str(tot_grd) + '</defaultgrade>' + '\n')
        f_ot.write('    <penalty>0.3333333</penalty>'                + '\n')
        # Penalty rate can be edited here.
        f_ot.write('    <hidden>0</hidden>'                          + '\n')
        f_ot.write('    <idnumber></idnumber>'                       + '\n')

        if   id_proc == 'singl':
            f_ot.write('    <single>true</single>'                       + '\n')
        elif id_proc == 'm_prt':
            f_ot.write('    <single>false</single>'                      + '\n')
        # single/multiple-choice control in Moodle.
        # this section is skipped if the case is all-or-nothing multiple choice
        # (id_proc == m_a_n)

        if   row[16] == '1':
            f_ot.write('    <shuffleanswers>true</shuffleanswers>'       + '\n')
        else:
            f_ot.write('    <shuffleanswers>false</shuffleanswers>'      + '\n')
        # shuffle answer choices.

        f_ot.write('    <answernumbering>abc</answernumbering>'      + '\n')
        f_ot.write('    <correctfeedback format=\"html\">'           + '\n')
        f_ot.write('      <text>あなたの答えは正解です。</text>'       + '\n')
        f_ot.write('    </correctfeedback>'                          + '\n')

        if id_proc == 'singl' or id_proc == 'm_prt':
            f_ot.write('    <partiallycorrectfeedback format=\"html\">'  + '\n')
            f_ot.write('      <text>あなたの答えは部分的に正解です。</text>'+ '\n')
            f_ot.write('    </partiallycorrectfeedback>'                 + '\n')
        # if the case is all-or-nothing multi-choice, this field is skipped
        # (id_proc == m_a_n).

        f_ot.write('    <incorrectfeedback format=\"html\">'         + '\n')
        f_ot.write('      <text>あなたの答えは正しくありません。</text>'+ '\n')
        f_ot.write('    </incorrectfeedback>'                        + '\n')

        if flag_nans == 1 and row[0] == '2':
            f_ot.write('    <shownumcorrect/>'                           + '\n')
        # show the number of correct answers after answering if the flag is set.
        # row[0]==2 means the question allows not single but multiple choices.

        #####
        ##### output xml for each answer choice
        #####

        i_c     = -1
        for c_now in choice:
            i_c += 1
            f_ot.write('    <answer fraction=\"' + str(c_grd[i_c]) + '\" format=\"html\">' + '\n')
            f_ot.write('      <text><![CDATA[<p>' + c_now + '<br></p>]]></text>' + '\n')
            f_ot.write('      <feedback format=\"html\">'                        + '\n')
            f_ot.write('        <text></text>'                                   + '\n')
            f_ot.write('      </feedback>'                                       + '\n')
            f_ot.write('    </answer>'                                           + '\n')
        # loop for each answer choice.

        f_ot.write('  </question>'                                           + '\n')
        f_ot.write('\n')

        id_cmp = 'yes'


    ##########
    ########## 穴埋め問題（文字列型）
    ##########

    if row[0] == '3':

        sntnce = row[6].split('\n')
        ans    = row[8].split('\n')
        # de-compose the question sentence and answers. This works!

        f_ot.write('<!-- question: row_' + str(i_rw) + '  -->'        + '\n')
        f_ot.write('  <question type=\"cloze\">'                      + '\n')
        f_ot.write('    <name>'                                       + '\n')
        f_ot.write('      <text>' + lab_q + '-' + row[2] + '</text>'  + '\n')
        f_ot.write('    </name>'                                      + '\n')
        f_ot.write('    <questiontext format=\"html\">'               + '\n')

        if flag_qttl == 1:   qttl = '<p><b>' + row[2] + '</b><br></p>'
        else:                qttl = ''
            # if flag_qttl is set, the title of the quetsion that was used
            # in cNavi is embedded at the beginning of the question text.

        #####
        ##### Grade (point)
        #####

        if row[5] == '1':
            grd = round( float(row[4])            )
        else:
            grd = round( float(row[4]) / len(ans) )
            if grd == 0:          grd  = int(1)
            # grd: point (Moodle says 'weight') for each cloze blank.
            # If row[5]==1 (partial point system) row[4] is given to each blank.
            # If row[5]!=1, row[4] is interpreted as total.

            # It seems Moodle accepts not float but only integer values for grades
            # of cloze questions. So 'round' is inserted to obtain grd when division
            # is used. grd could be zero then, which is corrected to grd=1.

        #####
        ##### Capital/small letter
        #####

        id_ltr = ['SHORTANSWER'] * len(ans)
            # id_ltr is [ SHORTANSWER / SHORTANSWER_C / NUMERICAL ] used in Moodle's cloze questions.
            # id_ltr is initialized here as SHORTANSWER.
            # SHORTANSWER_C is to distinguish Capital from small letters.
            # NUMERICAL     is to handle numbers.

        cn_ltr = row[24].split('\n')
            # cn_ltr is the cNavi setting recorded in csv:
            # [0: do not distinguish / 1: distinguish ] capital from small letters.
            # Note there should be some cases where the setting is incomplete
            # (e.g., empty, or num. of 0/1 specification is not the same as num. of answers)

        for i_l in range( len(id_ltr) ):
            if  i_l   <=  len(cn_ltr) - 1 :
                if   cn_ltr[i_l] == '1':
                     id_ltr[i_l]  = 'SHORTANSWER_C'
                elif cn_ltr[i_l] == '2':
                     id_ltr[i_l]  = 'NUMERICAL'
                     ans[i_l]     = ans[i_l].replace(' ', ':', 1).replace('m', '-')
                # If cNavi explicitly says '1' or '2' for a cloze blank, id_ltr is changed
                # to 'SHORTANSWER_C' and 'NUMERICAL', respectively.
                # - Note len(cn_ltr) should ideally be equal to len(ans), but there would be
                #   some irresponsible settings on csv, so 'if i_l <= len(cn_ltr) - 1'
                #   is inserted for safer treatment.
                # - As for NUMERICAL, csv stores a string '750 2' to mean 750+/-2.
                #   It is changed to '750:2' here by 'replace'.
                #   The reason the csv uses ' ' instead of ':' is because csv editors like
                #   Excel would transform '750:1', for example, to time automatically.
                # - One more thing, 'm' is used to mean '-' in csv files, because
                #   editors such as Excel would interpret minus sign as equation.
            else:
                f_lg.write('\n')
                f_lg.write(' Warning: The num. of Capital/Small indication in the csv insufficient.' + '\n')
                f_lg.write('          Line number   = ' + str(i_rw)                                  + '\n')
                # If the loop goes beyond the size of cNavi's indication (cn_ltr),
                # a warning is given.
                if        len(cn_ltr) > 0:
                    id_ltr[i_l]  =  id_ltr[ i_l - 1 ]
                # If len(cn_ltr)>0, csv has an indication on Capital/small letter treatment
                # at least for a cloze blank. In that case id_ltr for the others refers to the
                # last setting.


        #####
        ##### Zenkaku/Hankaku Control
        #####

        # The basic idea is to add zenkaku/hankaku versioon of the given answer
        # to correct answer as alternatives if Zenkaku/hankaku are not distinguished.

        id_zhc = [1] * len(ans)
            # id_zhc: Distinguish Zenkaku from Hankaku?
            # [ 0: No / 1: Yes ] (initialized as 1)
        cn_zhc = row[23].split('\n')
            # cn_zhc is the cNavi setting recorded in csv:
            # [0: do not distinguish / 1: distinguish ] Zenkaku from Hankaku.
            # Note there should be some cases where the setting is incomplete
            # (e.g., empty, or num. of 0/1 specification is not the same as num. of answers)

        for i_z in range( len(id_zhc) ):
            if  i_z   <=  len(cn_zhc) - 1 :
                if  cn_zhc[i_z] == '0':
                    id_zhc[i_z]  =  0
                # If cNavi says '0' explicitly for a cloze blank, id_zhc is changed to 0.
                # Note len(cn_zhc) should ideally be equal to len(ans), but there would be
                # some irresponsible settings on csv, so 'if i_z <= len(cn_zhc) - 1'
                # is inserted for safer treatment.
            else:
                f_lg.write('\n')
                f_lg.write(' Warning: The num. of Zenkaku/Hankaku indication in the csv insufficient.' + '\n')
                f_lg.write('          Line number   = ' + str(i_rw)                                    + '\n')
                # If the loop goes beyond the size of cNavi's indication (cn_zhc),
                # a warning is given.
                if        len(cn_zhc) > 0:
                    id_zhc[i_z]  =  id_zhc[ i_z - 1 ]
                # If len(cn_zhc)>0, csv has an indication on Zenkaku/Hankaku treatment
                # at least for a cloze blank. In that case id_zhc for the others refers to the
                # last setting.

                # Note.
                # The procedure taken here is the same as that for Capital/Small
                # letter treatment.

        an2 = [''] * len(ans)
                # an2: additonal answers for blanks to include zenkaku/hankaku answers.
                # This is initialized to empty.

        if flag_znhn == 1:
                # Case cNavi setting is to be transferred to Moodle.

            for i_z in range( len(id_zhc) ):

                if id_zhc[i_z] == 0:
                # Case Zenkaku/Hankaku are not distinguished from the other.
                # In that case, additional two answers with Zenkaku and Hankaku,
                # respectively, are added to an2 as right answers.
                # These could be added to ans directly, but a safer procedure is taken.

                    c_tmp_h = zen2han( ans[i_z] )
                    c_tmp_z = han2zen( ans[i_z] )
                    if c_tmp_h != ans[i_z]:    an2[i_z] = an2[i_z] + '~%100%' + c_tmp_h
                    if c_tmp_z != ans[i_z]:    an2[i_z] = an2[i_z] + '~%100%' + c_tmp_z
                    # c_tmp_h: the cNavi's answer but is transformed to Hankaku.
                    # c_tmp_z: the cNavi's answer but is transformed to Zenkaku.
                    # Note the answer strings could stay unchanged even after transformation.
                    # So they are added to an2 only when c_tmp_h/_z are different from the original answer.


        #####
        ##### Compose question sentence and answer
        #####

        q_txt = '<p>' + row[3] + '<br></p><p>'
            # The beginning part of the question text.

        i_s     = -1
        for a_now in ans:
            i_s += 1
            # Loop for each blank.

            ##### Width of cloze questions

            if flag_wdth == 1 and id_ltr[i_s] != 'NUMERICAL':   c_wdth = '~' + chr_rnd( int( float(row[17]) * fwdth ) )
            else:                                               c_wdth = ''
                # - Adjust the width of blanks. c_wdth is an empty or a random set
                #   of strings to be added as a 'false' (i.e., dummy) answer.
                #   Moodle does not have width control, so this program tries to
                #   control it by adding a dummy answer. chr_rnd is defined above.
                #   row[17] is the size of the cloze blank specified in cNavi csv.
                # - If the format is NUMERICAL, strings can not be accepted, so
                #   c_wdth is forced to set empty.

            ##### Question text, cloze blank, and the answer

            q_txt = q_txt + sntnce[i_s] + '{' + str(grd) + ':' + id_ltr[i_s] + ':=' + ans[i_s] + an2[i_s] + c_wdth + '}'
            # Add question sentence and answer for each blank to q_txt.
            # str(grd) is the grade of this cloze blank.
            # id_ltr is SHORTANSWER or SHORTANSWER_C depending on cNavi's setting.
            # ans is the correct answer.
            # an2 is the correct answer but in a different form, i.e., in Zenkaku/hankaku.
            # c_wdth is the random set of letters to adjust the blank size.

            ##### Close the question text

        i_s      += 1
        q_txt     = q_txt + sntnce[i_s] + '<br></p>'
            # Close the question sentence.


        #####
        ##### Output (The latter part)
        #####

        f_ot.write('      <text><![CDATA[' + qttl + q_txt + ']]></text>' + '\n')
        f_ot.write('    </questiontext>'                                 + '\n')
        f_ot.write('    <generalfeedback format=\"html\">'               + '\n')
        f_ot.write('      <text>' + row[13] + '</text>'                  + '\n')
        f_ot.write('    </generalfeedback>'                              + '\n')
        f_ot.write('    <penalty>0.3333333</penalty>'                    + '\n')
        # Penalty rate can be edited here.
        f_ot.write('    <hidden>0</hidden>'                              + '\n')
        f_ot.write('    <idnumber></idnumber>'                           + '\n')
        f_ot.write('  </question>'                                       + '\n')
        f_ot.write('\n')

        id_cmp = 'yes'



    ##########
    ########## check if the current row has been processed and completed
    ##########

    if id_cmp == 'no':
        f_lg.write('\n')
        f_lg.write(' Warning: The current line of the input csv was not processed.' + '\n')
        f_lg.write('          Line number   = ' + str(i_rw)                         + '\n')
        f_lg.write('          Question type = ' + row[0]                            + '\n')
        f_lg.write('          Question name = ' + row[2]                            + '\n')
        # If the current row of the input csv is left untouched,
        # its information is reported to the log file.


##########
########## End
##########

f_ot.write('</quiz>'                                                         + '\n')
    # Close <quiz> tag in the xml.

f_lg.write('\n')
f_lg.write(' Total number of rows in the input cNavi csv file: ' + str(i_rw) + '\n')

f_in.close()
f_ot.close()
f_lg.close()
